@@ -51,7 +51,9 @@ from .scope import (
     ColumnScope,
     IndicatorScope,
     ModelInputScope,
+    PendingOrderScope,
     PredictionScope,
+    PriceScope,
     StaticScope,
 )
 from collections import defaultdict, deque
@@ -63,6 +65,7 @@ from typing import (
     Callable,
     Iterable,
     Iterator,
+    Literal,
     Mapping,
     NamedTuple,
     Optional,
@@ -76,6 +79,11 @@ def _decorate_execution_fn(fn: Callable[[ExecContext], None]) -> Callable:
     ):
         set_exec_ctx_data(ctx, session, symbol, date)
         fn(ctx)
+        if ctx.buy_shares is not None and ctx.sell_shares is not None:
+            raise ValueError(
+                "For each symbol, only one of buy_shares or sell_shares can be"
+                " set per bar."
+            )
         if ctx.buy_limit_price is not None and ctx.buy_shares is None:
             raise ValueError(
                 "buy_shares must be set when buy_limit_price is set."
@@ -128,32 +136,6 @@ class Execution(NamedTuple):
     indicator_names: frozenset[str]
 
 
-class BacktestResult(NamedTuple):
-    r"""Result data from backtesting a :class:`.Strategy`.
-
-    Attributes:
-        start_date: Starting date of backtest.
-        end_date: Ending date of backtest.
-        portfolio_bars: ``deque`` of
-            :class:`pybroker.portfolio.PortfolioBar`\ s containing
-            :class:`pybroker.portfolio.Portfolio` data for every bar executed.
-        position_bars: ``deque`` of :class:`pybroker.portfolio.PositionBar`\ s
-            containing :class:`pybroker.portfolio.Position` data for every bar
-            executed.
-        orders: :class:`pybroker.portfolio.Order`\ s that were filled during
-            the backtest.
-        trades: :class:`pybroker.portfolio.Trade`\ s that were placed during
-            the backtest.
-    """
-
-    start_date: datetime
-    end_date: datetime
-    portfolio_bars: deque[PortfolioBar]
-    position_bars: deque[PositionBar]
-    orders: deque[Order]
-    trades: deque[Trade]
-
-
 class BacktestMixin:
     """Mixin implementing backtesting functionality."""
 
@@ -170,8 +152,15 @@ class BacktestMixin:
         max_long_positions: Optional[int],
         max_short_positions: Optional[int],
         pos_size_handler: Optional[Callable[[PosSizeContext], None]],
+        exit_dates: Mapping[str, np.datetime64],
+        exit_buy_fill_price: Union[
+            PriceType, Callable[[str, BarData], Union[int, float, Decimal]]
+        ] = PriceType.MIDDLE,
+        exit_sell_fill_price: Union[
+            PriceType, Callable[[str, BarData], Union[int, float, Decimal]]
+        ] = PriceType.MIDDLE,
         enable_fractional_shares: bool = False,
-    ) -> BacktestResult:
+    ):
         r"""Backtests a ``set`` of :class:`.Execution`\ s that implement
         trading logic.
 
@@ -197,6 +186,11 @@ class BacktestMixin:
                 held at a time. If ``None``, then unlimited.
             pos_size_handler: :class:`Callable` that sets position sizes when
                 placing orders for buy and sell signals.
+            exit_dates: :class:`Mapping` of symbols to exit dates.
+            exit_buy_fill_price: Fill price for covering an open short position
+                for ``exit_dates``.
+            exit_sell_fill_price: Fill price for selling an open long position
+                for ``exit_dates``.
             enable_fractional_shares: Whether to enable trading fractional
                 shares.
 
@@ -205,7 +199,6 @@ class BacktestMixin:
         """
         test_dates = test_data[DataCol.DATE.value].unique()
         test_dates.sort()
-        sym_exec_dates: dict[str, frozenset] = {}
         exec_symbols = tuple(
             ExecSymbol(exec.id, sym)
             for sym in test_data[DataCol.SYMBOL.value].unique()
@@ -228,15 +221,18 @@ class BacktestMixin:
         sell_sched: dict[np.datetime64, list[ExecResult]] = defaultdict(list)
         sym_end_index: dict[str, int] = defaultdict(int)
         col_scope = ColumnScope(test_data)
+        price_scope = PriceScope(col_scope, sym_end_index)
         ind_scope = IndicatorScope(indicator_data, test_dates)
         input_scope = ModelInputScope(col_scope, ind_scope)
         pred_scope = PredictionScope(models, input_scope)
+        pending_order_scope = PendingOrderScope()
         exec_ctx = ExecContext(
             portfolio=portfolio,
             col_scope=col_scope,
             ind_scope=ind_scope,
             input_scope=input_scope,
             pred_scope=pred_scope,
+            pending_order_scope=pending_order_scope,
             models=models,
             sym_end_index=sym_end_index,
         )
@@ -252,6 +248,7 @@ class BacktestMixin:
                 ind_scope=ind_scope,
                 input_scope=input_scope,
                 pred_scope=pred_scope,
+                pending_order_scope=pending_order_scope,
                 models=models,
                 sym_end_index=sym_end_index,
                 max_long_positions=max_long_positions,
@@ -261,6 +258,7 @@ class BacktestMixin:
         logger.backtest_executions_start(test_dates)
         buy_results: deque[ExecResult] = deque()
         sell_results: deque[ExecResult] = deque()
+        exit_syms: deque[str] = deque()
         for i, date in enumerate(test_dates):
             for sym in test_symbols:
                 if date not in sym_exec_dates[sym]:
@@ -288,29 +286,28 @@ class BacktestMixin:
             if is_sell_sched:
                 self._place_sell_orders(
                     date=date,
-                    df=test_data,
-                    col_scope=col_scope,
-                    sym_end_index=sym_end_index,
+                    price_scope=price_scope,
+                    pending_order_scope=pending_order_scope,
                     sell_sched=sell_sched,
-                    buy_sched=buy_sched,
                     portfolio=portfolio,
                     enable_fractional_shares=enable_fractional_shares,
                 )
             if is_buy_sched:
                 self._place_buy_orders(
                     date=date,
-                    df=test_data,
-                    col_scope=col_scope,
-                    sym_end_index=sym_end_index,
+                    price_scope=price_scope,
+                    pending_order_scope=pending_order_scope,
                     buy_sched=buy_sched,
-                    sell_sched=sell_sched,
                     portfolio=portfolio,
                     enable_fractional_shares=enable_fractional_shares,
                 )
+            portfolio.check_stops(date, price_scope)
             portfolio.capture_bar(date, test_data)
             for exec_symbol in exec_symbols:
                 if date not in sym_exec_dates[exec_symbol.symbol]:
                     continue
+                if exit_dates and date == exit_dates[exec_symbol.symbol]:
+                    exit_syms.append(exec_symbol.symbol)
                 result = exec_fns[exec_symbol.exec_id](
                     exec_ctx, sessions[exec_symbol], exec_symbol.symbol, date
                 )
@@ -323,29 +320,56 @@ class BacktestMixin:
             while buy_results:
                 self._schedule_order(
                     result=buy_results.popleft(),
+                    created=date,
                     sym_end_index=sym_end_index,
                     delay=buy_delay,
                     sched=buy_sched,
                     col_scope=col_scope,
+                    pending_order_scope=pending_order_scope,
                 )
             while sell_results:
                 self._schedule_order(
                     result=sell_results.popleft(),
+                    created=date,
                     sym_end_index=sym_end_index,
                     delay=sell_delay,
                     sched=sell_sched,
                     col_scope=col_scope,
+                    pending_order_scope=pending_order_scope,
+                )
+            while exit_syms:
+                self._exit_position(
+                    portfolio=portfolio,
+                    date=date,
+                    symbol=exit_syms.popleft(),
+                    exit_buy_fill_price=exit_buy_fill_price,
+                    exit_sell_fill_price=exit_sell_fill_price,
+                    price_scope=price_scope,
                 )
             portfolio.incr_bars()
             if i % 10 == 0 or i == len(test_dates) - 1:
                 logger.backtest_executions_loading(i + 1)
-        return BacktestResult(
-            start_date=to_datetime(test_dates[0]),
-            end_date=to_datetime(test_dates[-1]),
-            portfolio_bars=portfolio.bars,
-            position_bars=portfolio.position_bars,
-            orders=portfolio.orders,
-            trades=portfolio.trades,
+
+    def _exit_position(
+        self,
+        portfolio: Portfolio,
+        date: np.datetime64,
+        symbol: str,
+        exit_buy_fill_price: Union[
+            PriceType, Callable[[str, BarData], Union[int, float, Decimal]]
+        ],
+        exit_sell_fill_price: Union[
+            PriceType, Callable[[str, BarData], Union[int, float, Decimal]]
+        ],
+        price_scope: PriceScope,
+    ):
+        buy_fill_price = price_scope.fetch(symbol, exit_buy_fill_price)
+        sell_fill_price = price_scope.fetch(symbol, exit_sell_fill_price)
+        portfolio.exit_position(
+            date,
+            symbol,
+            buy_fill_price=buy_fill_price,
+            sell_fill_price=sell_fill_price,
         )
 
     def _set_pos_sizes(
@@ -366,17 +390,19 @@ class BacktestMixin:
                 if id >= (len(buy_results) + len(sell_results)):
                     raise ValueError(f"Invalid ExecSignal id: {id}")
                 if id < len(buy_results):
-                    buy_results[id].buy_shares = shares
+                    buy_results[id].buy_shares = to_decimal(shares)
                 else:
-                    sell_results[id - len(buy_results)].sell_shares = shares
+                    sell_results[
+                        id - len(buy_results)
+                    ].sell_shares = to_decimal(shares)
             elif buy_results is not None:
                 if id >= len(buy_results):
                     raise ValueError(f"Invalid ExecSignal id: {id}")
-                buy_results[id].buy_shares = shares
+                buy_results[id].buy_shares = to_decimal(shares)
             elif sell_results is not None:
                 if id >= len(sell_results):
                     raise ValueError(f"Invalid ExecSignal id: {id}")
-                sell_results[id].sell_shares = shares
+                sell_results[id].sell_shares = to_decimal(shares)
             else:
                 raise ValueError(
                     "buy_results and sell_results cannot both be None."
@@ -385,10 +411,12 @@ class BacktestMixin:
     def _schedule_order(
         self,
         result: ExecResult,
+        created: np.datetime64,
         sym_end_index: Mapping[str, int],
         delay: int,
         sched: Mapping[np.datetime64, list[ExecResult]],
         col_scope: ColumnScope,
+        pending_order_scope: PendingOrderScope,
     ):
         date_loc = sym_end_index[result.symbol] - 1
         dates = col_scope.fetch(result.symbol, DataCol.DATE.value)
@@ -397,6 +425,28 @@ class BacktestMixin:
         logger = StaticScope.instance().logger
         if date_loc + delay < len(dates):
             date = dates[date_loc + delay]
+            order_type: Literal["buy", "sell"]
+            if result.buy_shares is not None:
+                order_type = "buy"
+                shares = result.buy_shares
+                limit_price = result.buy_limit_price
+                fill_price = result.buy_fill_price
+            elif result.sell_shares is not None:
+                order_type = "sell"
+                shares = result.sell_shares
+                limit_price = result.sell_limit_price
+                fill_price = result.sell_fill_price
+            else:
+                raise ValueError("buy_shares or sell_shares needs to be set.")
+            result.pending_order_id = pending_order_scope.add(
+                type=order_type,
+                symbol=result.symbol,
+                created=created,
+                exec_date=date,
+                shares=shares,
+                limit_price=limit_price,
+                fill_price=fill_price,
+            )
             sched[date].append(result)
             logger.debug_schedule_order(date, result)
         else:
@@ -405,11 +455,9 @@ class BacktestMixin:
     def _place_buy_orders(
         self,
         date: np.datetime64,
-        df: pd.DataFrame,
-        col_scope: ColumnScope,
-        sym_end_index: Mapping[str, int],
+        price_scope: PriceScope,
+        pending_order_scope: PendingOrderScope,
         buy_sched: dict[np.datetime64, list[ExecResult]],
-        sell_sched: dict[np.datetime64, list[ExecResult]],
         portfolio: Portfolio,
         enable_fractional_shares: bool,
     ):
@@ -417,16 +465,17 @@ class BacktestMixin:
         for result in buy_results:
             if result.buy_shares is None:
                 continue
+            if (
+                result.pending_order_id is None
+                or not pending_order_scope.contains(result.pending_order_id)
+            ):
+                continue
+            pending_order_scope.remove(result.pending_order_id)
             buy_shares = self._get_shares(
                 result.buy_shares, enable_fractional_shares
             )
-            fill_price = self._get_price(
-                symbol=result.symbol,
-                date=date,
-                price=result.buy_fill_price,
-                df=df,
-                col_scope=col_scope,
-                end_index=sym_end_index[result.symbol],
+            fill_price = price_scope.fetch(
+                result.symbol, result.buy_fill_price
             )
             order = portfolio.buy(
                 date=date,
@@ -434,6 +483,7 @@ class BacktestMixin:
                 shares=buy_shares,
                 fill_price=fill_price,
                 limit_price=result.buy_limit_price,
+                stops=result.long_stops,
             )
             logger = StaticScope.instance().logger
             if order is None:
@@ -452,38 +502,14 @@ class BacktestMixin:
                     fill_price=fill_price,
                     limit_price=result.buy_limit_price,
                 )
-                if result.hold_bars is not None:
-                    if result.hold_bars <= 0:
-                        raise ValueError("hold_bars must be greater than 0.")
-                    sell_result = ExecResult(
-                        symbol=result.symbol,
-                        date=result.date,
-                        buy_fill_price=result.buy_fill_price,
-                        sell_shares=order.shares,
-                        sell_fill_price=result.sell_fill_price,
-                        sell_limit_price=result.sell_limit_price,
-                        score=result.score,
-                        hold_bars=None,
-                        buy_shares=None,
-                        buy_limit_price=None,
-                    )
-                    self._schedule_order(
-                        result=sell_result,
-                        sym_end_index=sym_end_index,
-                        delay=result.hold_bars,
-                        sched=sell_sched,
-                        col_scope=col_scope,
-                    )
         del buy_sched[date]
 
     def _place_sell_orders(
         self,
         date: np.datetime64,
-        df: pd.DataFrame,
-        col_scope: ColumnScope,
-        sym_end_index: Mapping[str, int],
+        price_scope: PriceScope,
+        pending_order_scope: PendingOrderScope,
         sell_sched: dict[np.datetime64, list[ExecResult]],
-        buy_sched: dict[np.datetime64, list[ExecResult]],
         portfolio: Portfolio,
         enable_fractional_shares: bool,
     ):
@@ -491,16 +517,17 @@ class BacktestMixin:
         for result in sell_results:
             if result.sell_shares is None:
                 continue
+            if (
+                result.pending_order_id is None
+                or not pending_order_scope.contains(result.pending_order_id)
+            ):
+                continue
+            pending_order_scope.remove(result.pending_order_id)
             sell_shares = self._get_shares(
                 result.sell_shares, enable_fractional_shares
             )
-            fill_price = self._get_price(
-                symbol=result.symbol,
-                date=date,
-                price=result.sell_fill_price,
-                df=df,
-                col_scope=col_scope,
-                end_index=sym_end_index[result.symbol],
+            fill_price = price_scope.fetch(
+                result.symbol, result.sell_fill_price
             )
             order = portfolio.sell(
                 date=date,
@@ -508,6 +535,7 @@ class BacktestMixin:
                 shares=sell_shares,
                 fill_price=fill_price,
                 limit_price=result.sell_limit_price,
+                stops=result.short_stops,
             )
             logger = StaticScope.instance().logger
             if order is None:
@@ -526,28 +554,6 @@ class BacktestMixin:
                     fill_price=fill_price,
                     limit_price=result.sell_limit_price,
                 )
-                if result.hold_bars is not None:
-                    if result.hold_bars <= 0:
-                        raise ValueError("hold_bars must be greater than 0.")
-                    buy_result = ExecResult(
-                        symbol=result.symbol,
-                        date=result.date,
-                        sell_fill_price=result.sell_fill_price,
-                        buy_shares=order.shares,
-                        buy_fill_price=result.buy_fill_price,
-                        buy_limit_price=result.buy_limit_price,
-                        score=result.score,
-                        hold_bars=None,
-                        sell_shares=None,
-                        sell_limit_price=None,
-                    )
-                    self._schedule_order(
-                        result=buy_result,
-                        sym_end_index=sym_end_index,
-                        delay=result.hold_bars,
-                        sched=buy_sched,
-                        col_scope=col_scope,
-                    )
         del sell_sched[date]
 
     def _get_shares(
@@ -560,49 +566,6 @@ class BacktestMixin:
         else:
             return to_decimal(int(shares))
 
-    def _get_price(
-        self,
-        symbol: str,
-        date: np.datetime64,
-        price: Union[
-            float,
-            int,
-            Decimal,
-            PriceType,
-            Callable[[BarData], Union[int, float, Decimal]],
-        ],
-        df: pd.DataFrame,
-        col_scope: ColumnScope,
-        end_index: int,
-    ) -> Decimal:
-        price_type = type(price)
-        if price_type == PriceType:
-            row = df.loc[(symbol, date)]
-            if price == PriceType.OPEN:
-                return to_decimal(row[DataCol.OPEN.value])
-            elif price == PriceType.HIGH:
-                return to_decimal(row[DataCol.HIGH.value])
-            elif price == PriceType.LOW:
-                return to_decimal(row[DataCol.LOW.value])
-            elif price == PriceType.CLOSE:
-                return to_decimal(row[DataCol.CLOSE.value])
-            elif price == PriceType.MIDDLE:
-                low = row[DataCol.LOW.value]
-                high = row[DataCol.HIGH.value]
-                return to_decimal(round((low + (high - low) / 2.0), 2))
-            elif price == PriceType.AVERAGE:
-                open_ = row[DataCol.OPEN.value]
-                high = row[DataCol.HIGH.value]
-                low = row[DataCol.LOW.value]
-                close = row[DataCol.CLOSE.value]
-                return to_decimal(round((open_ + low + high + close) / 4.0, 2))
-        if price_type == float or price_type == int or price_type == Decimal:
-            return to_decimal(price)  # type: ignore[arg-type]
-        if callable(price):
-            bar_data = col_scope.bar_data_from_data_columns(symbol, end_index)
-            return to_decimal(price(bar_data))
-        raise ValueError(f"Unknown price: {price_type}")
-
 
 class WalkforwardWindow(NamedTuple):
     """Contains ``train_data`` and ``test_data`` of a time window used for
@@ -1008,6 +971,7 @@ class Strategy(
                 - ``"h"``/``"hour"``: hours
                 - ``"d"``/``"day"``: days
                 - ``"w"``/``"week"``: weeks
+                - ``"mo"``/``"month"``: months
 
                 An example timeframe string is ``1h 30m``.
             between_time: ``tuple[str, str]`` of times of day e.g.
@@ -1092,6 +1056,7 @@ class Strategy(
                 - ``"h"``/``"hour"``: hours
                 - ``"d"``/``"day"``: days
                 - ``"w"``/``"week"``: weeks
+                - ``"mo"``/``"month"``: months
 
                 An example timeframe string is ``1h 30m``.
             between_time: ``tuple[str, str]`` of times of day e.g.
@@ -1235,12 +1200,20 @@ class Strategy(
         shuffle: bool,
         train_only: bool,
     ):
-        sessions: dict[ExecSymbol, dict] = {
-            ExecSymbol(execution.id, sym): {}
-            for execution in self._executions
-            if execution.fn is not None
-            for sym in execution.symbols
-        }
+        sessions: dict[ExecSymbol, dict] = {}
+        exit_dates: dict[str, np.datetime64] = {}
+        for exec in self._executions:
+            if exec.fn is None:
+                continue
+            for sym in exec.symbols:
+                sessions[ExecSymbol(exec.id, sym)] = {}
+                if self._config.exit_on_last_bar:
+                    sym_dates = df[df[DataCol.SYMBOL.value] == sym][
+                        DataCol.DATE.value
+                    ].values
+                    if len(sym_dates):
+                        sym_dates.sort()
+                        exit_dates[sym] = sym_dates[-1]
         for train_idx, test_idx in self.walkforward_split(
             df=df,
             windows=windows,
@@ -1287,7 +1260,10 @@ class Strategy(
                     max_long_positions=self._config.max_long_positions,
                     max_short_positions=self._config.max_short_positions,
                     pos_size_handler=self._pos_size_handler,
+                    exit_dates=exit_dates,
                     enable_fractional_shares=self._fractional_shares_enabled(),
+                    exit_buy_fill_price=self._config.exit_cover_fill_price,
+                    exit_sell_fill_price=self._config.exit_sell_fill_price,
                 )
 
     def _filter_dates(
