@@ -21,7 +21,7 @@ import re
 from .fixtures import *
 from collections import deque
 from decimal import Decimal
-from pybroker.common import PriceType, to_datetime
+from pybroker.common import PriceType, StopType, to_datetime
 from pybroker.context import (
     ExecContext,
     ExecResult,
@@ -29,7 +29,7 @@ from pybroker.context import (
     set_exec_ctx_data,
     set_pos_size_ctx_data,
 )
-from pybroker.portfolio import Order, Portfolio, Position, Trade
+from pybroker.portfolio import Order, Portfolio, Position, Stop, Trade
 
 
 @pytest.fixture()
@@ -104,6 +104,7 @@ def trades(dates, symbols):
         agg_pnl=Decimal(100),
         bars=1,
         pnl_per_bar=Decimal(100),
+        stop=None,
     )
 
 
@@ -113,6 +114,7 @@ def ctx(
     ind_scope,
     input_scope,
     pred_scope,
+    pending_order_scope,
     portfolio,
     trained_models,
     sym_end_index,
@@ -126,6 +128,7 @@ def ctx(
         ind_scope=ind_scope,
         input_scope=input_scope,
         pred_scope=pred_scope,
+        pending_order_scope=pending_order_scope,
         models=trained_models,
         sym_end_index=sym_end_index,
     )
@@ -139,6 +142,7 @@ def ctx_with_pos(
     ind_scope,
     input_scope,
     pred_scope,
+    pending_order_scope,
     portfolio,
     trained_models,
     sym_end_index,
@@ -159,6 +163,7 @@ def ctx_with_pos(
         ind_scope=ind_scope,
         input_scope=input_scope,
         pred_scope=pred_scope,
+        pending_order_scope=pending_order_scope,
         models=trained_models,
         sym_end_index=sym_end_index,
     )
@@ -172,6 +177,7 @@ def ctx_with_orders(
     ind_scope,
     input_scope,
     pred_scope,
+    pending_order_scope,
     portfolio,
     trained_models,
     sym_end_index,
@@ -189,6 +195,7 @@ def ctx_with_orders(
         ind_scope=ind_scope,
         input_scope=input_scope,
         pred_scope=pred_scope,
+        pending_order_scope=pending_order_scope,
         models=trained_models,
         sym_end_index=sym_end_index,
     )
@@ -460,9 +467,110 @@ def test_to_result(ctx, symbol, date):
         sell_limit_price=Decimal("110.11"),
         hold_bars=2,
         score=7,
+        long_stops=frozenset(
+            [
+                Stop(
+                    id=1,
+                    symbol=symbol,
+                    stop_type=StopType.BAR,
+                    pos_type="long",
+                    percent=None,
+                    points=None,
+                    bars=2,
+                    fill_price=PriceType.HIGH,
+                    limit_price=None,
+                )
+            ]
+        ),
+        short_stops=None,
     )
 
 
+@pytest.mark.parametrize("pos_type", ["long", "short"])
+@pytest.mark.parametrize(
+    "stop_attr, expected_stop_type",
+    [
+        ("stop_loss", StopType.LOSS),
+        ("stop_loss_pct", StopType.LOSS),
+        ("stop_profit", StopType.PROFIT),
+        ("stop_profit_pct", StopType.PROFIT),
+        ("stop_trailing", StopType.TRAILING),
+        ("stop_trailing_pct", StopType.TRAILING),
+    ],
+)
+def test_to_result_when_stop(
+    ctx, symbol, date, pos_type, stop_attr, expected_stop_type
+):
+    stop_limit = 200
+    stop_amount = 20
+    percent = None
+    points = None
+    if stop_attr.endswith("_pct"):
+        percent = stop_amount
+    else:
+        points = stop_amount
+    expected_stops = frozenset(
+        [
+            Stop(
+                id=1,
+                symbol=symbol,
+                stop_type=expected_stop_type,
+                pos_type=pos_type,
+                percent=percent,
+                points=points,
+                bars=None,
+                fill_price=None,
+                limit_price=stop_limit,
+            )
+        ]
+    )
+    buy_shares = None
+    sell_shares = None
+    long_stops = None
+    short_stops = None
+    if pos_type == "long":
+        buy_shares = 100
+        long_stops = expected_stops
+    else:
+        sell_shares = 100
+        short_stops = expected_stops
+    ctx.buy_shares = buy_shares
+    ctx.sell_shares = sell_shares
+    setattr(ctx, stop_attr, stop_amount)
+    setattr(ctx, f"{stop_attr.replace('_pct', '')}_limit", stop_limit)
+    result = ctx.to_result()
+    assert result == ExecResult(
+        symbol=symbol,
+        date=date,
+        buy_fill_price=PriceType.MIDDLE,
+        buy_shares=buy_shares,
+        buy_limit_price=None,
+        sell_fill_price=PriceType.MIDDLE,
+        sell_shares=sell_shares,
+        sell_limit_price=None,
+        hold_bars=None,
+        score=None,
+        long_stops=long_stops,
+        short_stops=short_stops,
+    )
+
+
+@pytest.mark.parametrize(
+    "stop_attr", ["stop_loss", "stop_profit", "stop_trailing"]
+)
+def test_to_result_when_stop_pct_and_points_then_error(ctx, stop_attr):
+    ctx.buy_shares = 100
+    with pytest.raises(
+        ValueError,
+        match=re.escape(
+            f"Only one of {stop_attr} or {stop_attr}_pct can be set."
+        ),
+    ):
+        setattr(ctx, stop_attr, 20)
+        setattr(ctx, f"{stop_attr}_pct", 20)
+        ctx.to_result()
+
+
 def test_orders(ctx_with_orders, orders):
     assert tuple(ctx_with_orders.orders()) == orders
 
@@ -483,11 +591,49 @@ def test_set_exec_ctx_data(ctx, symbols, sym_end_index):
     sym = symbols[-1]
     session = {"a": 1, "b": 2}
     date = np.datetime64("2020-01-01")
+    ctx._end_index = 1000
+    ctx._foreign = {"SPY": np.random.rand(100)}
+    ctx.buy_fill_price = PriceType.AVERAGE
+    ctx.buy_shares = 100
+    ctx.buy_limit_price = 99
+    ctx.sell_fill_price = PriceType.CLOSE
+    ctx.sell_shares = 200
+    ctx.sell_limit_price = 80
+    ctx.hold_bars = 5
+    ctx.score = 45.5
+    ctx.stop_loss = 10
+    ctx.stop_loss_pct = 20
+    ctx.stop_loss_limit = 99
+    ctx.stop_profit = 20
+    ctx.stop_profit_pct = 30
+    ctx.stop_profit_limit = 99.99
+    ctx.stop_trailing = 100
+    ctx.stop_trailing_pct = 15
+    ctx.stop_trailing_limit = 80.8
     set_exec_ctx_data(ctx, session, sym, date)
     assert ctx.session == session
     assert ctx.symbol == sym
     assert ctx.dt == to_datetime(date)
     assert ctx.bars == sym_end_index[sym]
+    assert not ctx._foreign
+    assert ctx._end_index == sym_end_index[sym]
+    assert ctx.buy_fill_price == PriceType.MIDDLE
+    assert ctx.buy_shares is None
+    assert ctx.buy_limit_price is None
+    assert ctx.sell_fill_price == PriceType.MIDDLE
+    assert ctx.sell_shares is None
+    assert ctx.sell_limit_price is None
+    assert ctx.hold_bars is None
+    assert ctx.score is None
+    assert ctx.stop_loss is None
+    assert ctx.stop_loss_pct is None
+    assert ctx.stop_loss_limit is None
+    assert ctx.stop_profit is None
+    assert ctx.stop_profit_pct is None
+    assert ctx.stop_profit_limit is None
+    assert ctx.stop_trailing is None
+    assert ctx.stop_trailing_pct is None
+    assert ctx.stop_trailing_limit is None
 
 
 def test_set_pos_ctx_data(
@@ -497,6 +643,7 @@ def test_set_pos_ctx_data(
     ind_scope,
     input_scope,
     pred_scope,
+    pending_order_scope,
     trained_models,
     sym_end_index,
 ):
@@ -512,6 +659,8 @@ def test_set_pos_ctx_data(
             sell_limit_price=None,
             hold_bars=None,
             score=1,
+            long_stops=None,
+            short_stops=None,
         ),
         ExecResult(
             symbol="AAPL",
@@ -524,6 +673,8 @@ def test_set_pos_ctx_data(
             sell_limit_price=None,
             hold_bars=None,
             score=2,
+            long_stops=None,
+            short_stops=None,
         ),
     ]
     sell_results = [
@@ -538,6 +689,8 @@ def test_set_pos_ctx_data(
             sell_limit_price=80,
             hold_bars=None,
             score=1,
+            long_stops=None,
+            short_stops=None,
         ),
     ]
     ctx = PosSizeContext(
@@ -546,6 +699,7 @@ def test_set_pos_ctx_data(
         ind_scope,
         input_scope,
         pred_scope,
+        pending_order_scope,
         trained_models,
         sym_end_index,
         max_long_positions=1,
@@ -582,3 +736,24 @@ def test_set_pos_ctx_data(
     assert all_signals[1].score == sell_signals[0].score
     assert all_signals[1].type == sell_signals[0].type
     assert all_signals[1].bar_data is not None
+
+
+def test_cancel_pending_order(ctx, pending_orders):
+    assert ctx.cancel_pending_order(pending_orders[0].id)
+    orders = tuple(ctx.pending_orders())
+    assert orders == tuple([pending_orders[1]])
+
+
+def test_cancel_all_pending_orders(ctx):
+    ctx.cancel_all_pending_orders()
+    assert not tuple(ctx.pending_orders())
+
+
+def test_cancel_all_pending_orders_when_symbol(ctx, pending_orders):
+    ctx.cancel_all_pending_orders("SPY")
+    assert tuple(ctx.pending_orders()) == tuple([pending_orders[1]])
+
+
+def test_pending_orders(ctx, pending_orders):
+    assert tuple(ctx.pending_orders()) == pending_orders
+    assert tuple(ctx.pending_orders("AAPL")) == tuple([pending_orders[1]])
