@@ -25,7 +25,7 @@ from .fixtures import *
 from collections import deque
 from datetime import datetime
 from decimal import Decimal
-from pybroker.common import DataCol, PriceType
+from pybroker.common import DataCol, PriceType, to_datetime
 from pybroker.config import StrategyConfig
 from pybroker.data import DataSource
 from pybroker.portfolio import (
@@ -35,6 +35,7 @@ from pybroker.portfolio import (
     PositionBar,
     Trade,
 )
+from pybroker.scope import PendingOrder
 from pybroker.strategy import (
     BacktestMixin,
     Execution,
@@ -43,7 +44,7 @@ from pybroker.strategy import (
     TestResult,
     WalkforwardMixin,
 )
-from unittest.mock import MagicMock, Mock
+from unittest.mock import Mock
 
 
 @pytest.fixture(params=[200, 202])
@@ -194,7 +195,7 @@ class TestBacktestMixin:
         execs = {buy_exec, sell_exec}
         mock_portfolio = Mock()
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
@@ -206,12 +207,8 @@ class TestBacktestMixin:
             max_long_positions=None,
             max_short_positions=None,
             pos_size_handler=pos_size_handler,
+            exit_dates={},
         )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
-        assert result.orders is not None
-        assert result.portfolio_bars is not None
-        assert result.position_bars is not None
         buy_df = data_source_df[data_source_df["symbol"] == "SPY"]
         buy_dates = buy_df["date"].unique()[1:]
         assert len(mock_portfolio.buy.call_args_list) == len(buy_dates)
@@ -253,7 +250,7 @@ class TestBacktestMixin:
         execs = {buy_exec}
         mock_portfolio = Mock()
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
@@ -265,12 +262,8 @@ class TestBacktestMixin:
             max_long_positions=None,
             max_short_positions=None,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
-        assert result.orders is not None
-        assert result.portfolio_bars is not None
-        assert result.position_bars is not None
         buy_df = data_source_df[data_source_df["symbol"] == "SPY"]
         buy_dates = buy_df["date"].unique()[2:]
         assert len(mock_portfolio.buy.call_args_list) == len(buy_dates)
@@ -300,7 +293,7 @@ class TestBacktestMixin:
         execs = {sell_exec}
         mock_portfolio = Mock()
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
@@ -312,12 +305,8 @@ class TestBacktestMixin:
             max_long_positions=None,
             max_short_positions=None,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
-        assert result.orders is not None
-        assert result.portfolio_bars is not None
-        assert result.position_bars is not None
         sell_df = data_source_df[data_source_df["symbol"] == "AAPL"]
         sell_dates = sell_df["date"].unique()[2:]
         assert len(mock_portfolio.sell.call_args_list) == len(sell_dates)
@@ -331,146 +320,6 @@ class TestBacktestMixin:
             )
             assert kwargs["limit_price"] == 50.5
 
-    def test_backtest_executions_when_buy_hold_bars(self, data_source_df):
-        def buy_exec_fn(ctx):
-            ctx.buy_fill_price = PriceType.CLOSE
-            ctx.sell_fill_price = PriceType.OPEN
-            ctx.buy_shares = 200
-            ctx.hold_bars = 2
-
-        buy_exec = Execution(
-            id=1,
-            symbols=frozenset(["SPY"]),
-            fn=buy_exec_fn,
-            model_names=frozenset(),
-            indicator_names=frozenset(),
-        )
-        execs = {buy_exec}
-        mock_portfolio = Mock()
-
-        def buy_fn(date, symbol, shares, fill_price, limit_price):
-            return Order(
-                id=1,
-                symbol=symbol,
-                type="buy",
-                date=date,
-                shares=shares,
-                fill_price=fill_price,
-                limit_price=limit_price,
-                fees=0,
-            )
-
-        mock_portfolio.buy = MagicMock(side_effect=buy_fn)
-        mixin = BacktestMixin()
-        result = mixin.backtest_executions(
-            executions=execs,
-            sessions=self._sessions(execs),
-            models={},
-            indicator_data={},
-            test_data=data_source_df,
-            portfolio=mock_portfolio,
-            buy_delay=1,
-            sell_delay=1,
-            max_long_positions=None,
-            max_short_positions=None,
-            pos_size_handler=None,
-        )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
-        df = data_source_df[data_source_df["symbol"] == "SPY"]
-        buy_dates = df["date"].unique()[1:]
-        assert len(mock_portfolio.buy.call_args_list) == len(buy_dates)
-        for i, date in enumerate(buy_dates):
-            _, kwargs = mock_portfolio.buy.call_args_list[i]
-            assert kwargs["date"] == date
-            assert kwargs["symbol"] == "SPY"
-            assert kwargs["shares"] == 200
-            assert kwargs["fill_price"] == Decimal(
-                str(df[df["date"] == date]["close"].values[0])
-            )
-            assert kwargs["limit_price"] is None
-        sell_dates = df["date"].unique()[3:]
-        assert len(mock_portfolio.sell.call_args_list) == len(sell_dates)
-        for i, date in enumerate(sell_dates):
-            _, kwargs = mock_portfolio.sell.call_args_list[i]
-            assert kwargs["date"] == date
-            assert kwargs["symbol"] == "SPY"
-            assert kwargs["shares"] == 200
-            assert kwargs["fill_price"] == Decimal(
-                str(df[df["date"] == date]["open"].values[0])
-            )
-            assert kwargs["limit_price"] is None
-
-    def test_backtest_executions_when_sell_hold_bars(self, data_source_df):
-        def sell_exec_fn(ctx):
-            ctx.sell_fill_price = PriceType.OPEN
-            ctx.buy_fill_price = PriceType.CLOSE
-            ctx.sell_shares = 100
-            ctx.hold_bars = 1
-
-        sell_exec = Execution(
-            id=1,
-            symbols=frozenset(["AAPL"]),
-            fn=sell_exec_fn,
-            model_names=frozenset(),
-            indicator_names=frozenset(),
-        )
-        execs = {sell_exec}
-        mock_portfolio = Mock()
-
-        def sell_fn(date, symbol, shares, fill_price, limit_price):
-            return Order(
-                id=1,
-                symbol=symbol,
-                type="sell",
-                date=date,
-                shares=shares,
-                fill_price=fill_price,
-                limit_price=limit_price,
-                fees=0,
-            )
-
-        mock_portfolio.sell = MagicMock(side_effect=sell_fn)
-        mixin = BacktestMixin()
-        result = mixin.backtest_executions(
-            executions=execs,
-            sessions=self._sessions(execs),
-            models={},
-            indicator_data={},
-            test_data=data_source_df,
-            portfolio=mock_portfolio,
-            buy_delay=1,
-            sell_delay=1,
-            max_long_positions=None,
-            max_short_positions=None,
-            pos_size_handler=None,
-        )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
-        df = data_source_df[data_source_df["symbol"] == "AAPL"]
-        sell_dates = df["date"].unique()[1:]
-        assert len(mock_portfolio.sell.call_args_list) == len(sell_dates)
-        for i, date in enumerate(sell_dates):
-            _, kwargs = mock_portfolio.sell.call_args_list[i]
-            assert kwargs["date"] == date
-            assert kwargs["symbol"] == "AAPL"
-            assert kwargs["shares"] == 100
-            assert kwargs["fill_price"] == Decimal(
-                str(df[df["date"] == date]["open"].values[0])
-            )
-            assert kwargs["limit_price"] is None
-        buy_dates = df["date"].unique()[2:]
-        assert len(mock_portfolio.buy.call_args_list) == len(buy_dates)
-        for i, date in enumerate(buy_dates):
-            _, kwargs = mock_portfolio.buy.call_args_list[i]
-            assert kwargs["date"] == date
-            assert kwargs["symbol"] == "AAPL"
-            assert kwargs["shares"] == 100
-            assert kwargs["fill_price"] == Decimal(
-                str(df[df["date"] == date]["close"].values[0])
-            )
-            assert kwargs["limit_price"] is None
-
     def test_backtest_executions_when_invalid_buy_hold_bars_then_error(
         self, data_source_df
     ):
@@ -502,6 +351,7 @@ class TestBacktestMixin:
                 max_long_positions=None,
                 max_short_positions=None,
                 pos_size_handler=None,
+                exit_dates={},
             )
 
     def test_backtest_executions_when_invalid_sell_hold_bars_then_error(
@@ -535,6 +385,7 @@ class TestBacktestMixin:
                 max_long_positions=None,
                 max_short_positions=None,
                 pos_size_handler=None,
+                exit_dates={},
             )
 
     def test_backtest_executions_when_no_fn(self, data_source_df):
@@ -546,25 +397,26 @@ class TestBacktestMixin:
             indicator_names=frozenset(),
         )
         execs = {exec}
+        portfolio = Portfolio(100_000)
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
             indicator_data={},
             test_data=data_source_df,
-            portfolio=Portfolio(100_000),
+            portfolio=portfolio,
             buy_delay=1,
             sell_delay=1,
             max_long_positions=None,
             max_short_positions=None,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
-        assert len(result.portfolio_bars)
-        assert not len(result.position_bars)
-        assert not len(result.orders)
+        assert len(portfolio.bars) == len(data_source_df["date"].unique())
+        assert not len(portfolio.position_bars)
+        assert not len(portfolio.orders)
+        assert not len(portfolio.trades)
 
     def test_backtest_executions_when_empty_symbols(self, data_source_df):
         def buy_exec_fn(ctx):
@@ -578,25 +430,26 @@ class TestBacktestMixin:
             indicator_names=frozenset(),
         )
         execs = {exec}
+        portfolio = Portfolio(100_000)
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
             indicator_data={},
             test_data=data_source_df[data_source_df["symbol"] != "AAPL"],
-            portfolio=Portfolio(100_000),
+            portfolio=portfolio,
             buy_delay=1,
             sell_delay=1,
             max_long_positions=None,
             max_short_positions=None,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
-        assert len(result.portfolio_bars)
-        assert not len(result.position_bars)
-        assert not len(result.orders)
+        assert len(portfolio.bars) == len(data_source_df["date"].unique())
+        assert not len(portfolio.position_bars)
+        assert not len(portfolio.orders)
+        assert not len(portfolio.trades)
 
     def test_backtest_executions_when_buy_delay_after_period(
         self, data_source_df
@@ -612,25 +465,27 @@ class TestBacktestMixin:
             indicator_names=frozenset(),
         )
         execs = {exec}
+        portfolio = Portfolio(100_000)
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
             indicator_data={},
             test_data=data_source_df,
-            portfolio=Portfolio(100_000),
+            portfolio=portfolio,
             buy_delay=1000,
             sell_delay=1,
             max_long_positions=None,
             max_short_positions=None,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
-        assert len(result.portfolio_bars)
-        assert not len(result.position_bars)
-        assert not len(result.orders)
+
+        assert len(portfolio.bars) == len(data_source_df["date"].unique())
+        assert not len(portfolio.position_bars)
+        assert not len(portfolio.orders)
+        assert not len(portfolio.trades)
 
     def test_backtest_executions_when_sell_delay_after_period(
         self, data_source_df
@@ -646,25 +501,26 @@ class TestBacktestMixin:
             indicator_names=frozenset(),
         )
         execs = {exec}
+        portfolio = Portfolio(100_000)
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
             indicator_data={},
             test_data=data_source_df,
-            portfolio=Portfolio(100_000),
+            portfolio=portfolio,
             buy_delay=1,
             sell_delay=1000,
             max_long_positions=None,
             max_short_positions=None,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
-        assert len(result.portfolio_bars)
-        assert not len(result.position_bars)
-        assert not len(result.orders)
+        assert len(portfolio.bars)
+        assert not len(portfolio.position_bars)
+        assert not len(portfolio.orders)
+        assert not len(portfolio.trades)
 
     def test_backtest_executions_when_buy_score(self, data_source_df):
         def buy_exec_fn(ctx):
@@ -685,7 +541,7 @@ class TestBacktestMixin:
         execs = {exec}
         mock_portfolio = Mock()
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
@@ -697,9 +553,8 @@ class TestBacktestMixin:
             max_long_positions=1,
             max_short_positions=None,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
         df = data_source_df[data_source_df["symbol"].isin(["AAPL", "SPY"])]
         buy_dates = sorted(df["date"].values)[2:]
         assert len(mock_portfolio.buy.call_args_list) == len(buy_dates)
@@ -737,7 +592,7 @@ class TestBacktestMixin:
         execs = {exec}
         mock_portfolio = Mock()
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
@@ -749,9 +604,8 @@ class TestBacktestMixin:
             max_long_positions=None,
             max_short_positions=1,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert result.start_date == data_source_df["date"].iloc[0]
-        assert result.end_date == data_source_df["date"].iloc[-1]
         df = data_source_df[data_source_df["symbol"].isin(["AAPL", "SPY"])]
         sell_dates = sorted(df["date"].values)[2:]
         assert len(mock_portfolio.sell.call_args_list) == len(sell_dates)
@@ -775,7 +629,7 @@ class TestBacktestMixin:
         [
             (50, 50),
             (Decimal(111.1), Decimal(111.1)),
-            (lambda _: 60, 60),
+            (lambda _symbol, _bar_data: 60, 60),
             (PriceType.OPEN, 200),
             (PriceType.HIGH, 400),
             (PriceType.LOW, 100),
@@ -814,7 +668,7 @@ class TestBacktestMixin:
         execs = {exec}
         mock_portfolio = Mock()
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
@@ -826,9 +680,8 @@ class TestBacktestMixin:
             max_long_positions=None,
             max_short_positions=None,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert result.start_date == dates[0]
-        assert result.end_date == dates[-1]
         buy_dates = dates[1:]
         assert len(mock_portfolio.buy.call_args_list) == len(buy_dates)
         for i, date in enumerate(buy_dates):
@@ -868,6 +721,7 @@ class TestBacktestMixin:
                 max_long_positions=None,
                 max_short_positions=None,
                 pos_size_handler=None,
+                exit_dates={},
             )
 
     def test_backtest_executions_when_buy_limit_and_no_shares_then_error(
@@ -903,6 +757,7 @@ class TestBacktestMixin:
                 max_long_positions=1,
                 max_short_positions=None,
                 pos_size_handler=None,
+                exit_dates={},
             )
 
     def test_backtest_executions_when_sell_limit_and_no_shares_then_error(
@@ -938,6 +793,7 @@ class TestBacktestMixin:
                 max_long_positions=1,
                 max_short_positions=None,
                 pos_size_handler=None,
+                exit_dates={},
             )
 
     def test_backtest_executions_when_buy_order_not_filled(
@@ -955,21 +811,23 @@ class TestBacktestMixin:
             indicator_names=frozenset(),
         )
         execs = {exec}
+        portfolio = Portfolio(1)
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
             indicator_data={},
             test_data=data_source_df,
-            portfolio=Portfolio(1),
+            portfolio=portfolio,
             buy_delay=1,
             sell_delay=1,
             max_long_positions=1,
             max_short_positions=None,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert not len(result.orders)
+        assert not len(portfolio.orders)
 
     def test_backtest_executions_when_sell_order_not_filled(
         self, data_source_df
@@ -987,21 +845,23 @@ class TestBacktestMixin:
             indicator_names=frozenset(),
         )
         execs = {exec}
+        portfolio = Portfolio(1)
         mixin = BacktestMixin()
-        result = mixin.backtest_executions(
+        mixin.backtest_executions(
             executions=execs,
             sessions=self._sessions(execs),
             models={},
             indicator_data={},
             test_data=data_source_df,
-            portfolio=Portfolio(1),
+            portfolio=portfolio,
             buy_delay=1,
             sell_delay=1,
             max_long_positions=1,
             max_short_positions=None,
             pos_size_handler=None,
+            exit_dates={},
         )
-        assert not len(result.orders)
+        assert not len(portfolio.orders)
 
     def _sessions(self, execs):
         return {
@@ -1187,17 +1047,11 @@ class TestStrategy:
                 START_DATE, "%Y-%m-%d"
             )
         else:
-            assert (
-                result.start_date
-                == pd.to_datetime(date_range[0]).to_pydatetime()
-            )
+            assert result.start_date == pd.to_datetime(date_range[0])
         if date_range[1] is None:
             assert result.end_date == datetime.strptime(END_DATE, "%Y-%m-%d")
         else:
-            assert (
-                result.end_date
-                == pd.to_datetime(date_range[1]).to_pydatetime()
-            )
+            assert result.end_date == pd.to_datetime(date_range[1])
         assert isinstance(result.portfolio, pd.DataFrame)
         assert not result.portfolio.empty
         assert isinstance(result.positions, pd.DataFrame)
@@ -1208,6 +1062,36 @@ class TestStrategy:
         else:
             assert result.bootstrap is None
 
+    def test_walkforward_results(self, data_source_df):
+        def exec_fn(ctx):
+            if not ctx.long_pos():
+                ctx.buy_shares = 100
+
+        data_source_df = data_source_df[
+            data_source_df["date"] <= to_datetime(END_DATE)
+        ]
+        strategy = Strategy(data_source_df, START_DATE, END_DATE)
+        strategy.add_execution(exec_fn, ["AAPL", "SPY"])
+        result = strategy.walkforward(windows=3, calc_bootstrap=False)
+        dates = set()
+        for _, test_idx in strategy.walkforward_split(
+            data_source_df, windows=3, lookahead=1, train_size=0.5
+        ):
+            df = data_source_df.loc[test_idx]
+            df = df[df["symbol"].isin(["AAPL", "SPY"])]
+            dates.update(df["date"].values)
+        assert result.start_date == to_datetime(START_DATE)
+        assert result.end_date == to_datetime(END_DATE)
+        dates_list = list(dates)
+        dates_list.sort()
+        assert np.array_equal(result.portfolio.index, dates_list)
+        assert len(result.positions) == 2 * len(dates) - 2
+        assert np.array_equal(
+            result.positions.index.get_level_values(1).unique(), dates_list[1:]
+        )
+        assert len(result.orders) == 2
+        assert not len(result.trades)
+
     def test_walkforward_when_no_executions_then_error(self, data_source_df):
         strategy = Strategy(data_source_df, START_DATE, END_DATE)
         with pytest.raises(
@@ -1606,6 +1490,7 @@ class TestStrategy:
                     agg_pnl=Decimal(1000),
                     bars=2,
                     pnl_per_bar=Decimal(500),
+                    stop=None,
                 ),
             )
         )
@@ -1655,3 +1540,404 @@ class TestStrategy:
         assert result.positions.empty
         assert result.orders.empty
         assert result.trades.empty
+
+    def test_backtest_when_exit_long_on_last_bar(self, data_source_df):
+        def buy_exec_fn(ctx):
+            if not ctx.long_pos():
+                ctx.buy_shares = 100
+                ctx.buy_fill_price = 150
+
+        def sell_fill_price(_symbol, _bar_data):
+            return 199.99
+
+        config = StrategyConfig(
+            exit_on_last_bar=True, exit_sell_fill_price=sell_fill_price
+        )
+        strategy = Strategy(data_source_df, START_DATE, END_DATE, config)
+        strategy.add_execution(buy_exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        dates = data_source_df[data_source_df["symbol"] == "SPY"][
+            "date"
+        ].unique()
+        dates = dates[dates <= np.datetime64(END_DATE)]
+        assert len(result.trades) == 1
+        trade = result.trades.iloc[0]
+        assert trade["type"] == "long"
+        assert trade["symbol"] == "SPY"
+        assert trade["entry_date"] == dates[1]
+        assert trade["exit_date"] == dates[-1]
+        assert trade["entry"] == 150
+        assert trade["exit"] == 199.99
+        assert trade["shares"] == 100
+
+    def test_backtest_when_exit_short_on_last_bar(self, data_source_df):
+        def sell_exec_fn(ctx):
+            if not ctx.short_pos():
+                ctx.sell_shares = 100
+                ctx.sell_fill_price = 200
+
+        def buy_fill_price(_symbol, _bar_data):
+            return 99.99
+
+        config = StrategyConfig(
+            exit_on_last_bar=True, exit_cover_fill_price=buy_fill_price
+        )
+        strategy = Strategy(data_source_df, START_DATE, END_DATE, config)
+        strategy.add_execution(sell_exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        dates = data_source_df[data_source_df["symbol"] == "SPY"][
+            "date"
+        ].unique()
+        dates = dates[dates <= np.datetime64(END_DATE)]
+        assert len(result.trades) == 1
+        trade = result.trades.iloc[0]
+        assert trade["type"] == "short"
+        assert trade["symbol"] == "SPY"
+        assert trade["entry_date"] == dates[1]
+        assert trade["exit_date"] == dates[-1]
+        assert trade["entry"] == 200
+        assert trade["exit"] == 99.99
+        assert trade["shares"] == 100
+
+    def test_backtest_when_buy_shares_and_sell_shares_then_error(
+        self, data_source_df
+    ):
+        def exec_fn(ctx):
+            ctx.buy_shares = 100
+            ctx.sell_shares = 100
+
+        strategy = Strategy(data_source_df, START_DATE, END_DATE)
+        strategy.add_execution(exec_fn, ["AAPL", "SPY"])
+        with pytest.raises(
+            ValueError,
+            match=re.escape(
+                "For each symbol, only one of buy_shares or sell_shares can be"
+                " set per bar."
+            ),
+        ):
+            strategy.backtest()
+
+    def test_backtest_pending_orders(self, data_source_df):
+        buy_delay = 2
+        dates = data_source_df[data_source_df["symbol"] == "SPY"][
+            "date"
+        ].unique()
+        dates = dates[dates <= np.datetime64(END_DATE)]
+
+        def buy_exec_fn(ctx):
+            if ctx.bars == 1:
+                ctx.buy_shares = 100
+            elif ctx.bars == 2:
+                orders = tuple(ctx.pending_orders())
+                assert len(orders) == 1
+                assert orders[0] == PendingOrder(
+                    id=1,
+                    type="buy",
+                    symbol="SPY",
+                    created=ctx.date[0],
+                    exec_date=dates[buy_delay],
+                    shares=100,
+                    limit_price=None,
+                    fill_price=PriceType.MIDDLE,
+                )
+            else:
+                assert not tuple(ctx.pending_orders())
+
+        config = StrategyConfig(buy_delay=buy_delay)
+        strategy = Strategy(data_source_df, START_DATE, END_DATE, config)
+        strategy.add_execution(buy_exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        assert len(result.orders) == 1
+        order = result.orders.iloc[0]
+        assert order["type"] == "buy"
+        assert order["symbol"] == "SPY"
+        assert order["date"] == dates[2]
+        assert np.isnan(order["limit_price"])
+        assert order["shares"] == 100
+
+    def test_backtest_when_pending_orders_canceled(self, data_source_df):
+        dates = data_source_df[data_source_df["symbol"] == "SPY"][
+            "date"
+        ].unique()
+        dates = dates[dates <= np.datetime64(END_DATE)]
+        buy_delay = 10
+        sell_delay = 5
+
+        def exec_fn(ctx):
+            if ctx.bars == 1:
+                ctx.buy_shares = 100
+                ctx.buy_limit_price = 99
+            elif ctx.bars == 2:
+                ctx.sell_shares = 200
+                ctx.sell_limit_price = 100
+            elif ctx.bars == 3:
+                orders = tuple(ctx.pending_orders())
+                assert len(orders) == 2
+                assert orders[0] == PendingOrder(
+                    id=1,
+                    type="buy",
+                    symbol="SPY",
+                    created=ctx.date[0],
+                    exec_date=dates[buy_delay],
+                    shares=100,
+                    limit_price=99,
+                    fill_price=PriceType.MIDDLE,
+                )
+                assert orders[1] == PendingOrder(
+                    id=2,
+                    type="sell",
+                    symbol="SPY",
+                    created=ctx.date[1],
+                    exec_date=dates[1 + sell_delay],
+                    shares=200,
+                    limit_price=100,
+                    fill_price=PriceType.MIDDLE,
+                )
+                ctx.cancel_all_pending_orders()
+            else:
+                assert not tuple(ctx.pending_orders())
+
+        config = StrategyConfig(buy_delay=buy_delay, sell_delay=sell_delay)
+        strategy = Strategy(data_source_df, START_DATE, END_DATE, config)
+        strategy.add_execution(exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        assert not len(result.orders)
+
+    def test_backtest_when_buy_hold_bars(self, data_source_df):
+        def buy_exec_fn(ctx):
+            ctx.buy_fill_price = PriceType.CLOSE
+            ctx.sell_fill_price = PriceType.OPEN
+            ctx.buy_shares = 100
+            ctx.hold_bars = 2
+
+        df = data_source_df[data_source_df["symbol"] == "SPY"]
+        dates = df["date"].unique()
+        dates = dates[dates <= np.datetime64(END_DATE)]
+        buy_dates = dates[1:]
+        sell_dates = dates[3:]
+        config = StrategyConfig(initial_cash=500_000)
+        strategy = Strategy(data_source_df, START_DATE, END_DATE, config)
+        strategy.add_execution(buy_exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        orders = result.orders
+        buy_orders = orders[orders["type"] == "buy"]
+        assert len(buy_orders) == len(buy_dates)
+        for buy_date in buy_dates:
+            row = buy_orders[buy_orders["date"] == buy_date]
+            assert row["symbol"].item() == "SPY"
+            assert row["shares"].item() == 100
+            assert np.isnan(row["limit_price"].item())
+            assert row["fill_price"].item() == round(
+                df[df["date"] == buy_date]["close"].item(), 2
+            )
+            assert row["fees"].item() == 0
+        sell_orders = orders[orders["type"] == "sell"]
+        assert len(sell_orders) == len(sell_dates)
+        for sell_date in sell_dates:
+            row = sell_orders[sell_orders["date"] == sell_date]
+            assert row["symbol"].item() == "SPY"
+            assert row["shares"].item() == 100
+            assert np.isnan(row["limit_price"].item())
+            assert row["fill_price"].item() == round(
+                df[df["date"] == sell_date]["open"].item(), 2
+            )
+            assert row["fees"].item() == 0
+        assert (result.trades["stop"] == "bar").all()
+
+    def test_backtest_when_sell_hold_bars(self, data_source_df):
+        def sell_exec_fn(ctx):
+            ctx.sell_fill_price = PriceType.OPEN
+            ctx.buy_fill_price = PriceType.CLOSE
+            ctx.sell_shares = 100
+            ctx.hold_bars = 1
+
+        df = data_source_df[data_source_df["symbol"] == "SPY"]
+        dates = df["date"].unique()
+        dates = dates[dates <= np.datetime64(END_DATE)]
+        buy_dates = dates[2:]
+        sell_dates = dates[1:]
+        strategy = Strategy(data_source_df, START_DATE, END_DATE)
+        strategy.add_execution(sell_exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        orders = result.orders
+        sell_orders = orders[orders["type"] == "sell"]
+        assert len(sell_orders) == len(sell_dates)
+        for sell_date in sell_dates:
+            row = sell_orders[sell_orders["date"] == sell_date]
+            assert row["symbol"].item() == "SPY"
+            assert row["shares"].item() == 100
+            assert np.isnan(row["limit_price"].item())
+            assert row["fill_price"].item() == round(
+                df[df["date"] == sell_date]["open"].item(), 2
+            )
+            assert row["fees"].item() == 0
+        buy_orders = orders[orders["type"] == "buy"]
+        assert len(buy_orders) == len(buy_dates)
+        for buy_date in buy_dates:
+            row = buy_orders[buy_orders["date"] == buy_date]
+            assert row["symbol"].item() == "SPY"
+            assert row["shares"].item() == 100
+            assert np.isnan(row["limit_price"].item())
+            assert row["fill_price"].item() == round(
+                df[df["date"] == buy_date]["close"].item(), 2
+            )
+            assert row["fees"].item() == 0
+        assert len(result.trades) == len(buy_orders)
+        assert (result.trades["stop"] == "bar").all()
+
+    def test_backtest_when_stop_loss(self, data_source_df):
+        def exec_fn(ctx):
+            if ctx.bars == 1:
+                ctx.buy_shares = 100
+                ctx.stop_loss = 10
+
+        df = data_source_df[data_source_df["symbol"] == "SPY"]
+        dates = df["date"].unique()
+        dates = dates[dates <= np.datetime64(END_DATE)]
+        strategy = Strategy(data_source_df, START_DATE, END_DATE)
+        strategy.add_execution(exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        assert len(result.trades) == 1
+        trade = result.trades.iloc[0]
+        assert trade["type"] == "long"
+        assert trade["symbol"] == "SPY"
+        assert trade["entry_date"] == dates[1]
+        assert trade["exit"] == trade["entry"] - 10
+        assert trade["shares"] == 100
+        assert trade["pnl"] == -1000
+        assert trade["agg_pnl"] == -1000
+        assert trade["pnl_per_bar"] == round(-1000 / trade["bars"], 2)
+        assert trade["stop"] == "loss"
+        assert len(result.orders) == 2
+        buy_order = result.orders.iloc[0]
+        assert buy_order["type"] == "buy"
+        assert buy_order["symbol"] == "SPY"
+        assert buy_order["date"] == dates[1]
+        assert buy_order["shares"] == 100
+        assert np.isnan(buy_order["limit_price"])
+        assert buy_order["fees"] == 0
+        sell_order = result.orders.iloc[1]
+        assert sell_order["type"] == "sell"
+        assert sell_order["symbol"] == "SPY"
+        assert sell_order["shares"] == 100
+        assert np.isnan(sell_order["limit_price"])
+        assert sell_order["fees"] == 0
+
+    def test_backtest_when_sell_before_stop_loss(self, data_source_df):
+        def exec_fn(ctx):
+            if ctx.bars == 1:
+                ctx.buy_shares = 100
+                ctx.stop_loss = 10
+            elif ctx.bars == 10:
+                ctx.sell_all_shares()
+
+        df = data_source_df[data_source_df["symbol"] == "SPY"]
+        dates = df["date"].unique()
+        dates = dates[dates <= np.datetime64(END_DATE)]
+        strategy = Strategy(data_source_df, START_DATE, END_DATE)
+        strategy.add_execution(exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        assert len(result.trades) == 1
+        trade = result.trades.iloc[0]
+        assert trade["type"] == "long"
+        assert trade["symbol"] == "SPY"
+        assert trade["entry_date"] == dates[1]
+        assert trade["exit_date"] == dates[10]
+        assert trade["shares"] == 100
+        assert trade["stop"] is None
+        assert len(result.orders) == 2
+        buy_order = result.orders.iloc[0]
+        assert buy_order["type"] == "buy"
+        assert buy_order["symbol"] == "SPY"
+        assert buy_order["date"] == dates[1]
+        assert buy_order["shares"] == 100
+        assert np.isnan(buy_order["limit_price"])
+        assert buy_order["fees"] == 0
+        sell_order = result.orders.iloc[1]
+        assert sell_order["type"] == "sell"
+        assert sell_order["symbol"] == "SPY"
+        assert sell_order["date"] == dates[10]
+        assert sell_order["shares"] == 100
+        assert np.isnan(sell_order["limit_price"])
+        assert sell_order["fees"] == 0
+
+    def test_backtest_when_cancel_stop(self, data_source_df):
+        def exec_fn(ctx):
+            if ctx.bars == 1:
+                ctx.buy_shares = 100
+                ctx.stop_loss = 10
+            elif ctx.bars == 10:
+                assert ctx.cancel_stop(stop_id=1)
+
+        df = data_source_df[data_source_df["symbol"] == "SPY"]
+        dates = df["date"].unique()
+        dates = dates[dates <= np.datetime64(END_DATE)]
+        strategy = Strategy(data_source_df, START_DATE, END_DATE)
+        strategy.add_execution(exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        assert not len(result.trades)
+        assert len(result.orders) == 1
+        buy_order = result.orders.iloc[0]
+        assert buy_order["type"] == "buy"
+        assert buy_order["symbol"] == "SPY"
+        assert buy_order["date"] == dates[1]
+        assert buy_order["shares"] == 100
+        assert np.isnan(buy_order["limit_price"])
+        assert buy_order["fees"] == 0
+
+    def test_backtest_when_cancel_stops(self, data_source_df):
+        def exec_fn(ctx):
+            if ctx.bars == 1:
+                ctx.buy_shares = 100
+                ctx.stop_loss = 10
+                ctx.stop_trailing = 10
+            elif ctx.bars == 10:
+                ctx.cancel_stops("SPY")
+
+        df = data_source_df[data_source_df["symbol"] == "SPY"]
+        dates = df["date"].unique()
+        dates = dates[dates <= np.datetime64(END_DATE)]
+        strategy = Strategy(data_source_df, START_DATE, END_DATE)
+        strategy.add_execution(exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        assert not len(result.trades)
+        assert len(result.orders) == 1
+        buy_order = result.orders.iloc[0]
+        assert buy_order["type"] == "buy"
+        assert buy_order["symbol"] == "SPY"
+        assert buy_order["date"] == dates[1]
+        assert buy_order["shares"] == 100
+        assert np.isnan(buy_order["limit_price"])
+        assert buy_order["fees"] == 0
+
+    def test_backtest_when_pos_size_handler_zero_shares(self, data_source_df):
+        def buy_exec_fn(ctx):
+            ctx.buy_shares = 100
+
+        def sell_exec_fn(ctx):
+            ctx.sell_shares = 100
+
+        def pos_size_handler(ctx):
+            signals = tuple(ctx.signals())
+            ctx.set_shares(signals[0], shares=0)
+            ctx.set_shares(signals[1], shares=0)
+
+        strategy = Strategy(data_source_df, START_DATE, END_DATE)
+        strategy.add_execution(buy_exec_fn, "SPY")
+        strategy.add_execution(sell_exec_fn, "AAPL")
+        strategy.set_pos_size_handler(pos_size_handler)
+        result = strategy.backtest(calc_bootstrap=False)
+        assert not len(result.orders)
+
+    def test_backtest_when_no_stops(self, data_source_df):
+        def exec_fn(ctx):
+            if ctx.bars == 1:
+                ctx.buy_shares = 100
+            elif ctx.bars > 30:
+                ctx.sell_all_shares()
+
+        strategy = Strategy(data_source_df, START_DATE, END_DATE)
+        strategy.add_execution(exec_fn, "SPY")
+        result = strategy.backtest(calc_bootstrap=False)
+        assert len(result.trades) == 1
+        assert result.trades.iloc[0]["stop"] is None
